-- wengwengweng

local Node = Object:extend()

Node.name = "Node"
Node.unique = true
Node.deps = {}
Node.index = 0

function Node:new(...)

	self.parent = nil
	self.root = nil
	self._tags = {}
	self._args = {...}
	self._children = {}
	self._drawing = true
	self._updating = true
	self._listening = true
	self._input = Input()

end

--[[ ### ---------
	checking if a node is suitable for parent
--------- ### ]]

function Node:_check(p)

	if (self.unique) then
		if (p[self.__index]) then
			error("unique node")
		end
	end

	for i, d in ipairs(self.deps) do
		if (not p[d]) then
			error("dependency not met")
		end
	end

end

--[[ ### ---------
	add a child node
--------- ### ]]

function Node:n(n)

	if (not Object.is(n, Node)) then
		error("not a node")
	end

	if (n.parent) then
		error("has parent")
	end

	n:_check(self)
	n.parent = self
	n.root = self.root
	n:init(unpack(n._args))
	n:input(n._input)

	local index = n.__index

	if (self[index]) then

		if (n.unique) then
			error("unique node")
		else
			self[index][#self[index] + 1] = n
		end

	else

		if (n.unique) then
			self[index] = n
		else
			self[index] = { n }
		end

	end

	self._children[#self._children + 1] = n
	self:_sort()

	return n

end

--[[ ### ---------
	adopt a child node
--------- ### ]]

function Node:adopt(n)

	if (not Object.is(n, Node)) then
		error("not a node")
	end

	if (not n.parent) then
		error("doesn't have parent")
	end

	n:_check(self)
	n.parent:remove(n)
	n.parent = self
	n.root = self.root

	local index = n.__index

	if (self[index]) then

		if (n.unique) then
			error("unique node")
		else
			self[index][#self[index] + 1] = n
		end

	else

		if (n.unique) then
			self[index] = n
		else
			self[index] = { n }
		end

	end

	self._children[#self._children + 1] = n
	self:_sort()

	return n

end

--[[ ### ---------
	sort child nodes
--------- ### ]]

function Node:_sort()

	local swapped = true

	while (swapped) do

		swapped = false

		for i = 2, #self._children do
			if (self._children[i - 1].index > self._children[i].index) then
				self._children[i - 1], self._children[i] = self._children[i], self._children[i - 1]
				swapped = true
			end
		end

	end

end

--[[ ### ---------
	remove a child node
--------- ### ]]

function Node:remove(thing)

	local l = {}

	if (Object.is(thing, Node)) then

		local index = thing.__index

		if (thing.unique) then

			self[index] = nil
			_.remove(self._children, thing)

			l = { thing }

		else

			if (_.remove(self[index], thing) and _.remove(self._children, thing)) then

				if (#self[index] == 0) then
					self[index] = nil
				end

				l = { thing }

			end

		end

	elseif (self[thing]) then

		self[thing] = nil

		for i = #self._children, 1, -1 do
			if (Object.is(self._children[i], thing)) then
				l[#l + 1] = self._children[i]
				table.remove(self._children, i)
			end
		end

	end

	return l

end

--[[ ### ---------
	destroy a child node
--------- ### ]]

function Node:destroy(c)

	local l = self:remove(c)

	if (l) then
		for i, n in ipairs(l) do
			n:collect()
		end
	end

end

function Node:setDraw(b)
	self._drawing = b
end

function Node:setUpdate(b)
	self._updating = b
end

function Node:setListen(b)
	self._listening = b
end

function Node:z(i)

	self.index = i
	self.parent:_sort()

	return self

end

function Node:t(...)

	for i, t in ipairs({...}) do
		self._tags[#self._tags + 1] = t
	end

end

function Node:hasTag(t)

	return _.has(self._tags, t)

end

function Node:get(t)

	local l = {}

	for i, n in ipairs(self._children) do
		if (n:hasTag(t)) then
			l[#l + 1] = n
		end
	end

	return l

end

function Node:listen(dt)

	self._input:update(dt)

	local l = self._children

	for i = #l, 1, -1 do
		if (l[i]._listening) then
			l[i]:listen(dt)
		end
	end

end

function Node:update(dt)

	local l = self._children

	for k, n in ipairs(self._children) do
		if (n._updating) then
			n:update(dt)
		end
	end

end

function Node:draw()

	local l = self._children

	love.graphics.push()

	for k, n in ipairs(self._children) do
		if (n._drawing) then
			n:draw()
		end
	end

	love.graphics.pop()

end

function Node:collect()

	for i, t in ipairs(self._children) do
		t:collect()
	end

end

function Node:init() end
function Node:input() end
function Node:enter() end
function Node:resume() end
function Node:leave() end

return Node
