-- wengwengweng

local Shader = Object:extend()

local list = {

	vhs = {

		code = [[
			float a = 0.0025 * intensity;
			final.r = texture2D(tex, vec2(tc.x + a * dirs[0], tc.y + a * dirs[1])).r;
			final.g = texture2D(tex, vec2(tc.x + a * dirs[2], tc.y + a * dirs[3])).g;
			final.b = texture2D(tex, vec2(tc.x + a * dirs[4], tc.y + a * dirs[5])).b;
		]], externs = { intensity = 3.0, dirs = { 1.0, 0.0, 0.0, 1.0, -1.0, -1.0 } }

	},

	invert = {

		code = [[
			final.r = 1.0 - final.r;
			final.g = 1.0 - final.g;
			final.b = 1.0 - final.b;
		]], externs = {}

	},

	noise = {

		code = [[
			tc -= rand(tc, rnd) / (1000.0 / intensity);
			final = Texel(tex, tc);
		]], externs = { intensity = 10.0, rnd = 1.0 }, funcs = { "rand" }

	},

	tvnoise = {

		code = [[
			final.r -= -light + 1 + rand(tc, rnd);
			final.g -= -light + 1 + rand(tc, rnd);
			final.b -= -light + 1 + rand(tc, rnd);
		]], externs = { light = 1, rnd = 1 }, funcs = { "rand" }

	},

	distortion = {

		code = [[
			final.r = (sin(final.r * amount) + 1.0) * 0.5;
			final.g = (sin(final.g * amount) + 1.0) * 0.5;
			final.b = (sin(final.b * amount) + 1.0) * 0.5;
		]], externs = { amount = 20 }, funcs = { "wave" }

	},

	waves = {

		code = [[
			tc.x += sin(tc.y * curves + time * 6.0) * (intensity * 0.03);
			final = Texel(tex, tc);
		]], externs = { time = 0.0, curves = 5.0, intensity = 1.0 }

	},

	hdr = {

		code = [[
			vec4 intens = smoothstep(0.2, 0.8, final) + normalize(vec4(final.xyz, 1.0));
			if (fract(sc.y * 0.5) > 0.5) {
				intens = final * 0.8;
			}
			intens.a = 1.0;
			final = intens;
		]], externs = {}

	}

}

local functions = {

	rand = [[
		float rand(vec2 co, float v) {
			return fract(sin(dot(co.xy, vec2(12.9898, 78.233))) * 43758.5453 * v);
		}
	]],

	wave = [[
		float wave(float x, float amount) {
			return (sin(x * amount) + 1.0) * 0.5;
		}
	]]

}

function Shader:new(name)

	if (not list[name]) then
		return
	end

	local e = list[name]

	local function getExterns(externs)

		local function declare(name, val)

			local t = type(val)

			if (t == "number") then
				return "float " .. name
			end

			if (t == "boolean") then
				return "bool " .. name
			end

			if (t == "table") then
				local t2 = type(val[1])
				if t2 == "number" then
					if #val <= 4 then
						return "vec" .. #val .. " " .. name
					else
						return "float " .. name .. "[" .. #val .. "]"
					end
				end
			end

		end

		local text = {}

		for name, val in pairs(externs) do
			text[#text + 1] = "extern " .. declare(name, val) .. ";\n"
		end

		return table.concat(text, "")

	end

	local function getFuncs(funcs)

		local text = {}

		for _, f in ipairs(funcs) do
			text[#text + 1] = functions[f] or ""
		end

		return table.concat(text, "")

	end

	local externs = e.externs and getExterns(e.externs) or ""
	local funcs = e.funcs and getFuncs(e.funcs) or ""

	local header = [[
		vec4 effect(vec4 color, Image tex, vec2 tc, vec2 sc) {
			vec4 final = Texel(tex, tc);
	]]

	local footer = [[
			return final;
		}
	]]

	self.code = externs .. funcs .. header .. e.code .. footer
	self.effect = love.graphics.newShader(self.code)

	for name, val in pairs(e.externs) do
		if (type(val) == "table") then
			self:send(name, unpack(val))
		else
			self:send(name, val)
		end
	end

end

function Shader:push()
	love.graphics.setShader(self.effect)
end

function Shader:send(...)
	self.effect:send(...)
end

function Shader:pop()
	love.graphics.setShader()
end

return Shader
