-- wengwengweng

local Box = Node:extend()

Box.name = "Box"
Box.unique = false
Box.deps = { T }

function Box:init(v, w)

	self.vertices = v or {}
	self.angle = 0
	self.scale = 1
	self.solid = false
	self.tags = {}

	if (w) then
		self:setWorld(w)
	end

end

function Box:collides(other)
	return col.sat(self:getGlobalVertices(), other:getGlobalVertices())
end

function Box:hasPoint(x, y)
	return col.collidePointPoly(x, y, self:getGlobalVertices())
end

function Box:intersects(x1, y1, x2, y2)
	return col.collideLinePoly(x1, y1, x2, y2, self:getGlobalVertices())
end

function Box:setWorld(w)

	assert(Object.is(w, World), "not a valid world")

	if (self.world) then
		self:remove()
	end

	self.world = w
	self:register()

	return self

end

function Box:setSolid(s)

	self.solid = s

	return self

end

function Box:addTags(...)

	for i, t in ipairs({...}) do
		self.tags[#self.tags + 1] = t
	end

	return self

end

function Box:hasTag(tag)

	for i, t in ipairs(self.tags) do
		if (t == tag) then
			return true
		end
	end

	return false

end

function Box:map(f)

	assert(#self.vertices >= 3, "please add vertices first")

	local x, y = col.getPolyCenter(self.vertices)

	for i, v in ipairs(self.vertices) do
		f(v, x, y)
	end

end

function Box:setAngle(r)

	local diff = r - self.angle

	self:map(function(v, x, y)

		local dis = _.distance(x, y, v[1], v[2])
		local angle = _.angle(x, y, v[1], v[2])
		local dx, dy = _.vector(angle + math.rad(diff), dis)

		v[1] = x + dx
		v[2] = y + dy

	end)

	self.angle = self.angle + diff

	return self

end

function Box:setScale(s)

	local diff = s / self.scale

	self:map(function(v, x, y)

		local dis = _.distance(x, y, v[1], v[2])
		local angle = _.angle(x, y, v[1], v[2])
		local dx, dy = _.vector(angle, dis * diff)

		v[1] = x + dx
		v[2] = y + dy

	end)

	self.scale = self.scale * diff

	return self

end

function Box:setVertices(v)

	assert(type(v) == "table", "not valid vertices")
	assert(#v >= 3, "must have more than 3 vertices")
	self.vertices = v

	return self

end

function Box:getGlobalVertices()

	local vertices = {}
	local gp = self.parent[T]:global()

	for i, v in ipairs(self.vertices) do
		vertices[i] = { v[1] + gp.pos.x, v[2] + gp.pos.y }
	end

	return vertices

end

function Box:move(dx, dy)

	assert(self.world, "doesn't have a world")

	local list = {}

	self.parent.pos:add(dx, dy)

	self.world:map(function(b)

		if (b ~= self) then

			local collision, overlap, axis = self:collides(b)

			if (collision) then

				list[#list + 1] = b

				if (b.solid) then
					self.parent.pos:sub(_.vector(_.angle(0, 0, dx, dy), overlap))
				end

			end

		end

	end)

	return list

end

function Box:getCollisions()

	assert(self.world, "doesn't have a world")
	assert(#self.vertices >= 3, "please add vertices first")

	local cols = {}

	self.world:map(function(b)
		if (b ~= self) then
			if (self:collides(b)) then
				cols[#cols + 1] = b
			end
		end
	end)

	return cols

end

function Box:register()

	assert(self.world, "doesn't have a world")
	self.world:add(self)
	assert(self.world:has(self), "failed to add box")

	return self

end

function Box:collect()

	if (self.world) then
		self.world:remove(self)
	end

	return self

end

function Box:resolve(f)

	assert(self.world, "doesn't have a world")

	for i, c in ipairs(self:getCollisions()) do
		f(c.parent)
	end

end

function Box:debug()

	if (#self.vertices < 3) then
		return
	end

	if (not self.parent._drawing) then
		return
	end

	love.graphics.setColor({ 1, 1, 1 })
	love.graphics.setLineWidth(2)
	love.graphics.polygon("line", unpack(_.concat(unpack(self:getGlobalVertices()))))

end

return Box
